
"""
.. module:: test_basic_BVE.py
    :synopsis: It contains the scripts used to test the functions of
    module_basic_BVE.py
.. moduleauthor:: Kereman Alonso (alonso.kerman5@gmail.com)
"""

import numpy as np
from numpy import float64, arcsin
from math import sqrt as sqrt, factorial as fac, pi, cos,\
    asin, exp
from cmath import exp as cexp
import numba
from scipy.special import roots_legendre, lpmn
from scipy.fftpack import fft, ifft

from matplotlib import pyplot as plt, cm, colors
from mpl_toolkits.mplot3d import Axes3D

from netCDF4 import Dataset

import get_external_measurements as ext_measurement
from module_basic_BVE import *

if __name__ == "__main__":

    """
    Test legendre polinomials, laplacian, and longitudinal derivative
    """
    # # Calculate associated legendre polinomials
    # # M: triangular truncation parameter
    # M = 33
    # latit_dim = 1
    # latitudes = np.array([-0.5], dtype=np.float64)
    # legendre_matrix = np.zeros((M+2, 2*(M+1)+1, latitudes.shape[0]),
    #                            dtype=np.float64)
    # beta_matrix = np.ones((M+2, M+2), dtype=np.float64)
    # associated_legendre(latitudes, legendre_matrix, beta_matrix)
    # print('Values of associated legendre polinomials:')
    # print(legendre_matrix[33, 0+(33), 0])

    # print('Values of beta:')
    # print(beta_matrix[15, 10])

    # # Calculate longitudinal partial derivative of function knowing its
    # # spectral coeficients
    # spectral_coef = np.ones((M+1, 2*M+1), dtype=np.complex)
    # longit_partial_matrix = np.ones((M+1, 2*M+1), dtype=np.complex)
    # longit_partial(spectral_coef, longit_partial_matrix)
    # print('Spectral coeficients of the longitudinal partial derivative:')
    # print(longit_partial_matrix)

    # # Calculate laplacian of function knowing its
    # # spectral coeficients
    # laplace_matrix = np.ones((M+1, 2*M+1), dtype=np.complex)
    # laplacean(spectral_coef, 1, laplace_matrix)
    # print('Spectral coeficients of the laplacean:')
    # print(laplace_matrix)

    """
    Test latitudinal derivative
    """
    # M = 33
    # longit_dim = 3 * M + 1
    # latit_dim = (3 * M + 1) // 2 if (3 * M + 1) % 2 == 0 else\
    #     (3 * M + 1) // 2 + 1
    # longitudes, latitudes, weights = coordinates_and_weights(
    #     longit_dim, latit_dim)

    # legendre_matrix = np.zeros((M+2, 2*(M+1)+1, latitudes.shape[0]),
    #                            dtype=np.float64)
    # beta_matrix = np.ones((M+2, M+2), dtype=np.float64)
    # associated_legendre(latitudes, legendre_matrix, beta_matrix)

    # f_grid_values = np.zeros((latit_dim, longit_dim), dtype=np.float64)
    # f_spectral_coef = np.zeros((M+1, 2*M+1), dtype=np.complex)

    # # Enter initial values of the grid points: mu**3
    # for latit in range(latit_dim):
    #     for longit in range(longit_dim):
    #         f_grid_values[-latit+latit_dim-1, longit] = latitudes[latit]**3

    # deriv_grid_values = np.zeros((latit_dim, longit_dim), dtype=np.float64)
    # deriv_spectral_coef = np.zeros((M+2, 2*(M+1)+1), dtype=np.complex)

    # forward_transform(f_grid_values, weights, legendre_matrix, f_spectral_coef)

    # # Compute partial derivative in coeficients
    # quasi_latit_partial(f_spectral_coef, beta_matrix, deriv_spectral_coef)

    # inverse_transform(deriv_spectral_coef, legendre_matrix,
    #                   longitudes, deriv_grid_values)

    # # Calculate the results analitically: (mu**2-1)*(3*mu**2)
    # analitical_deriv = np.zeros((latit_dim, longit_dim), dtype=np.float64)
    # for latit in range(latit_dim):
    #     for longit in range(longit_dim):
    #         analitical_deriv[-latit+latit_dim-1, longit] =\
    #             (latitudes[latit]**2 - 1) * (3 * latitudes[latit]**2)

    # # Compute the difference between the two results
    # for latit in range(latit_dim):
    #     for longit in range(longit_dim):
    #         dif = deriv_grid_values[-latit+latit_dim-1, longit] -\
    #             analitical_deriv[-latit+latit_dim-1, longit]
    #         if (dif > 1e-4):
    #             print('Dif too big:', dif)

    """
    (7.5) Ex. 7.5 using spectral method with spherical harmonics.
    """
    # # omega: angular velocity of Earth in rad/s.
    # omega = 7.2921e-5
    # # a: radius of Earth in metres.
    # a = 6.3781e6

    # # All time unis are in seconds.
    # t_step = 900
    # forecast_step = 900  # * 4 * 24
    # max_time = 72 * 3600

    # # M: truncation parameter.
    # M = 33
    # longit_dim = 3 * M + 1
    # latit_dim = (3 * M + 1) // 2 if (3 * M + 1) % 2 == 0 else\
    #     (3 * M + 1) // 2 + 1

    # # Set initial values of streamfunction
    # A = -5.0e6
    # B = 7.4e8
    # C = -6.0e7
    # D = 5.0e7
    # sigma1 = 3.5
    # sigma2 = 6.0

    # psi_grid = np.zeros((latit_dim, longit_dim), dtype=float64)
    # for latit in range(latit_dim):
    #     for longit in range(longit_dim):
    #         r1 = sqrt((longit - 70.0) ** 2 + (latit - 20.0) ** 2)
    #         r2 = sqrt((longit - 80.0) ** 2 + (latit - 30.0) ** 2)
    #         psi_grid[-latit+latit_dim-1, longit] = A * latit + B + C *\
    #             exp(- (r1 ** 2) / (2 * sigma1 ** 2))\
    #             + D * exp(- (r2 ** 2) / (2 * sigma2 ** 2))

    # # Plot the initial values of streamfunction
    # latitudes, weights = roots_legendre(latit_dim)
    # latitudes_plot = - arcsin(latitudes)
    # longit_step = 2 * pi / longit_dim
    # longitudes = np.array([longit_step * i - pi for i in range(longit_dim)],
    #                       dtype=np.float64)

    # plt.figure(1)
    # plt.contour(longitudes, latitudes_plot, psi_grid, levels=10,
    #             colors='black')
    # print(0, psi_grid[-2+latit_dim+1, 2])

    # # Color map
    # # # Normalize vorticity values
    # norm = cm.colors.Normalize(vmin=psi_grid.min(),
    #                             vmax=psi_grid.max())
    # fig = plt.figure(2)
    # color_plot = plt.pcolormesh(longitudes, latitudes_plot, psi_grid,
    #                             cmap=cm.hot,
    #                             norm=norm)
    # fig.colorbar(color_plot)
    # plt.show()

    # # Get the values of the associated Legendre polinomials and the values
    # # of beta.
    # legendre_matrix = np.zeros((M+2, 2*(M+1)+1, latit_dim), dtype=np.float64)
    # beta_matrix = np.zeros((M+2, M+2), dtype=float64)
    # associated_legendre(latitudes, legendre_matrix, beta_matrix)

    """
    (7.5) Test if forward and inverse transform work.
    """

    # # Make forward transform to get the spectral coeficients of psi_grid
    # # named psi_coef_past
    # psi_coef_past = np.zeros((M+1, 2*M+1), dtype=np.complex)

    # re_psi_grid = np.zeros((latit_dim, longit_dim), dtype=float64)
    # forward_transform(psi_grid, weights, legendre_matrix,
    #                   psi_coef_past)
    # inverse_transform(psi_coef_past, legendre_matrix,
    #                   longitudes, re_psi_grid)
    # print(np.array([psi_grid / re_psi_grid], dtype=float64).max())
    # # for i in range(100):
    # #     forward_transform(re_psi_grid, weights, legendre_matrix,
    # #                       psi_coef_past)
    # #     inverse_transform(psi_coef_past, legendre_matrix,
    # #                       longitudes, psi_grid)
    # #     print(2, psi_grid[-2+latit_dim+1, 2])

    # # Plot psi_grid
    # latitudes_plot = - arcsin(latitudes)
    # plt.figure(3)
    # plt.contour(longitudes, latitudes_plot, psi_grid, levels=10,
    #             colors='black')

    # # Color map
    # # # Normalize vorticity values
    # norm = cm.colors.Normalize(vmin=re_psi_grid.min(),
    #                             vmax=re_psi_grid.max())
    # fig = plt.figure(4)
    # color_plot = plt.pcolormesh(longitudes, latitudes_plot, re_psi_grid,
    #                             cmap=cm.hot,
    #                             norm=norm)
    # fig.colorbar(color_plot)
    # plt.show()

    """
    (7.5) Test forecast.
    """
    # # Integrte forward in time
    # psi_forecast(t_step, forecast_step, max_time, omega, a, psi_grid, M)

    """
    (ERA5) Use the initial conditions gotten from ERA5.
    """
    # omega: angular velocity of Earth in rad/s.
    omega = 7.2921e-5
    # a: radius of Earth in metres.
    a = 6.3781e6
    
    year = '2019'
    month = '07'
    # Get the information stored in the nc file
    an = Dataset('..\\..\\data\\an_{}_{}.nc'.format(year, month), 'r')


    # t0: index of initial time in ERA Interim analysis time array.
    # t0 = 0 corresponds to the first day of a month at 00:00
    day = 10
    t0 = (day - 1) * 4
    # pressure: index of pressure in ERA5 pressure array. pressure = 2
    # corresponds to 500 hPa
    pressure = 2
    
    # order: order of antialiasing
    order = 2
    M = 63  # M = 479 gives highest possible mesh resolution

    longit_dim = 3 * M + 1
    latit_dim = (3 * M + 1) // 2 if (3 * M + 1) % 2 == 0 else\
        (3 * M + 1) // 2 + 1
    longitudes, latitudes, weights =\
        coordinates_and_weights(longit_dim, latit_dim)
    latitudes_plot = -arcsin(latitudes)

    # f: planetary vorticity
    f = np.zeros(latit_dim, dtype=float64)
    get_planetary_vo(omega, latitudes, f)

    vorticity = np.zeros((latit_dim, longit_dim), dtype=float64)

    ext_measurement.get_vorticity_matrix(an, t0, pressure, latitudes,
                                     longitudes, order, vorticity)

    psi = np.zeros((latit_dim, longit_dim), dtype=float64)
    
    # Plot initial vorticity
    # # Color map
    # # # Normalize vorticity values
    norm = cm.colors.Normalize(vmin=vorticity.min()*1e5, vmax=vorticity.max()*1e5)
    fig = plt.figure(1, dpi=150)
    color_plot = plt.pcolormesh(longitudes, latitudes_plot, vorticity*1e5,
                                cmap=cm.hot,
                                norm=norm)
    fig.colorbar(color_plot)
    fig.suptitle('Time: {}h'.format(0))
    plt.show()

    """
    (ERA5) Test forward and inverse transforms
    """
    # re_vorticity = np.zeros((latit_dim, longit_dim), dtype=float64)

    # legendre_matrix = np.zeros((M+2, 2*(M+1)+1, latit_dim), dtype=np.float64)
    # beta_matrix = np.zeros((M+2, M+2), dtype=float64)
    # associated_legendre(latitudes, legendre_matrix, beta_matrix)

    # vo_coef = np.zeros((M+1, 2*M+1), dtype=np.complex)
    # forward_transform(vorticity, weights, legendre_matrix, vo_coef)
    # inverse_transform(vo_coef, legendre_matrix, longitudes, re_vorticity)

    # re_vorticity_old = np.zeros((latit_dim, longit_dim), dtype=float64)
    # for i in range(10):
    #     re_vorticity_old[:] = re_vorticity[:]
    #     forward_transform(re_vorticity_old, weights, legendre_matrix, vo_coef)
    #     inverse_transform(vo_coef, legendre_matrix, longitudes, re_vorticity)

    # # Plot vorticity after transforms
    # # # Color map
    # # # # Normalize vorticity values
    # norm = cm.colors.Normalize(vmin=vorticity.min(),
    #                             vmax=vorticity.max())
    # fig = plt.figure(2)
    # color_plot = plt.pcolormesh(longitudes, latitudes_plot, re_vorticity,
    #                             cmap=cm.hot,
    #                             norm=norm)
    # fig.colorbar(color_plot)
    # plt.show()

    # print(np.average(np.array([vorticity / re_vorticity], dtype=float64)))
    # print(vorticity[25, 50] - re_vorticity[25, 50])

    """
    (ERA5) Test vorticity_to_psi and psi_to_vorticity.
    """
    # t_vorticity = np.zeros((latit_dim, longit_dim), dtype=float64)
    # psi = np.zeros((latit_dim, longit_dim), dtype=float64)
    # re_vorticity = np.zeros((latit_dim, longit_dim), dtype=float64)

    # legendre_matrix = np.zeros((M+2, 2*(M+1)+1, latit_dim), dtype=np.float64)
    # beta_matrix = np.zeros((M+2, M+2), dtype=float64)
    # associated_legendre(latitudes, legendre_matrix, beta_matrix)

    # # Make forward and inverse transform of vorticity to get the vorticity
    # # that can be represented with M modes
    # vo_coef = np.zeros((M+1, 2*M+1), dtype=np.complex)
    # forward_transform(vorticity, weights, legendre_matrix, vo_coef)
    # inverse_transform(vo_coef, legendre_matrix, longitudes, t_vorticity)

    # # # Plot truncated vorticity
    # norm = cm.colors.Normalize(vmin=vorticity.min(),
    #                            vmax=vorticity.max())
    # fig = plt.figure(3)
    # color_plot = plt.pcolormesh(longitudes, latitudes_plot, t_vorticity,
    #                             cmap=cm.hot,
    #                             norm=norm)
    # fig.colorbar(color_plot)
    # plt.show()

    # # vorticity_to_psi
    # vorticity_to_psi(t_vorticity, longitudes, latitudes, weights, a, M, psi)

    # # psi_to_vorticity
    # psi_to_vorticity(psi, longitudes, latitudes, weights, a, M, re_vorticity)

    # # Compare initial vorticity to vorticity after vo_to_psi and psi_to_vo
    # # transformations.

    # # # Plot truncated vorticity after vo_to_psi and psi_to_vo
    # # # transformations.
    # norm = cm.colors.Normalize(vmin=vorticity.min(),
    #                            vmax=vorticity.max())
    # fig = plt.figure(4)
    # color_plot = plt.pcolormesh(longitudes, latitudes_plot, re_vorticity,
    #                             cmap=cm.hot,
    #                             norm=norm)
    # fig.colorbar(color_plot)
    # plt.show()

    # print('Average ratio (not very indicative):')
    # print(np.average(np.array([t_vorticity / re_vorticity], dtype=float64)))

    # print('Maximum relative error:')
    # print(np.abs(np.array((t_vorticity - re_vorticity) /
    #                       t_vorticity, dtype=float64)).max())

    # print('Indice of maximum relative error point:')
    # flat = np.argmax(np.abs(np.array((t_vorticity - re_vorticity) /
    #                         t_vorticity, dtype=float64)))
    # indices = (flat // longit_dim,
    #            flat % longit_dim)
    # print(indices)

    # print('Average relative error')
    # print(np.average(np.abs(np.array((t_vorticity - re_vorticity) /
    #                                  t_vorticity,
    #                                  dtype=float64))))

    """
    (ERA5) Test psi_to_u and psi_to_v.
    """
    t_vorticity = np.zeros((latit_dim, longit_dim), dtype=float64)
    u_nc = np.zeros((latit_dim, longit_dim), dtype=float64)
    v_nc = np.zeros((latit_dim, longit_dim), dtype=float64)
    t_u = np.zeros((latit_dim, longit_dim), dtype=float64)
    t_v = np.zeros((latit_dim, longit_dim), dtype=float64)
    psi = np.zeros((latit_dim, longit_dim), dtype=float64)
    u = np.zeros((latit_dim, longit_dim), dtype=float64)
    v = np.zeros((latit_dim, longit_dim), dtype=float64)

    legendre_matrix = np.zeros((M+2, 2*(M+1)+1, latit_dim), dtype=np.float64)
    beta_matrix = np.zeros((M+2, M+2), dtype=float64)
    associated_legendre(latitudes, legendre_matrix, beta_matrix)

    ext_measurement.get_u_matrix(an, t0, pressure, latitudes, longitudes, 
                                 order,
                                 u_nc)
    norm_u = cm.colors.Normalize(vmin=u_nc.min(),
                                 vmax=u_nc.max())
    ext_measurement.get_v_matrix(an, t0, pressure, latitudes, longitudes, 
                                 order,
                                 v_nc)
    norm_v = cm.colors.Normalize(vmin=v_nc.min(),
                                 vmax=v_nc.max())

    # Make forward and inverse transforms
    u_coef = np.zeros((M+2, 2*(M+1)+1), dtype=np.complex)
    forward_transform(u_nc, weights, legendre_matrix, u_coef)
    inverse_transform(u_coef, legendre_matrix, longitudes, t_u)
    v_coef = np.zeros((M+1, 2*M+1), dtype=np.complex)
    forward_transform(v_nc, weights, legendre_matrix, v_coef)
    inverse_transform(v_coef, legendre_matrix, longitudes, t_v)
    vo_coef = np.zeros((M+1, 2*M+1), dtype=np.complex)
    forward_transform(vorticity, weights, legendre_matrix, vo_coef)
    inverse_transform(vo_coef, legendre_matrix, longitudes, t_vorticity)


    # # Plot truncated initial u
    fig = plt.figure(4, dpi=150)
    color_plot = plt.pcolormesh(longitudes, latitudes_plot, t_u,
                                cmap=cm.Reds,
                                norm=norm_u)
    fig.colorbar(color_plot)
    fig.suptitle('u nc')
    plt.show()

    # # Plot truncated initial v
    fig = plt.figure(5, dpi=150)
    color_plot = plt.pcolormesh(longitudes, latitudes_plot, t_v,
                                cmap=cm.Blues,
                                norm=norm_v)
    fig.colorbar(color_plot)
    fig.suptitle('v nc')
    plt.show()

    # vorticity_to_psi
    vorticity_to_psi(t_vorticity, longitudes, latitudes, weights, a, M, psi)

    # psi to u
    psi_to_u(psi, longitudes, latitudes, weights, a, M, u)
    
    # Plot u
    fig = plt.figure(100, dpi=150)
    color_plot = plt.pcolormesh(longitudes, latitudes_plot, u,
                                cmap=cm.Reds,
                                norm=norm_u)
    fig.colorbar(color_plot)
    fig.suptitle('u from psi'.format(0))
    plt.show()
    
    # psi to v
    psi_to_v(psi, longitudes, latitudes, weights, a, M, v)
    
    # Plot v
    fig = plt.figure(101, dpi=150)
    color_plot = plt.pcolormesh(longitudes, latitudes_plot, v,
                                cmap=cm.Blues,
                                norm=norm_v)
    fig.colorbar(color_plot)
    fig.suptitle('v from psi'.format(0))
    plt.show()
    
    # Plot arrow map
    fig = plt.figure(102, dpi=150)
    arrow_plot = plt.quiver(
        longitudes[180:191]*180/pi+180-360, latitudes_plot[20:31]*180/pi, 
                            u[20:31, 180:191], v[20:31, 180:191],
                            color='b',
                            headwidth=3,
                            headlength=5,
                            pivot='mid' )
    # plt.plot([-3.0153, -2.9679], [43.3803, 43.3419], color='r')
    plt.show()

